// Copyright © 2020 Graciliano M. P. All rights reserved.
// This code is governed by the Apache License, Version 2.0.
// Please refer to the LICENSE and AUTHORS files for details.

import 'dart:typed_data';

import 'package:data_serializer/data_serializer.dart';

/// Base class for code generation output.
mixin GeneratedOutput<O extends Object, T extends Object> {
  void writeAll(Iterable<T> list) {
    for (var e in list) {
      write(e);
    }
  }

  void write(T o);

  O output();
}

/// Generated Bytes.
/// - [toString] will show bytes description.
class BytesOutput extends BytesEmitter
    implements GeneratedOutput<Uint8List, List<int>> {
  BytesOutput({super.data, super.description});
}
